#include <unistd.h>
#include <signal.h>

#include <map>

#include <Process.h>
#include <Socket.h>
#include <Exception.h>
#include <Poller.h>


//----------------------------------------------------------------------------
struct ClientEntry
{
    TCPConnection *connection;
    std::string nick;
};


//----------------------------------------------------------------------------
class Server : public Process, public PollerCallBackInterface
{
    typedef std::map<int, ClientEntry> ClientMap;

  public:
    Server();
    ~Server();

    int run();

  protected:

    void shutdown();

    void onPoll(int fd, pollevent_t events);
    void onTerm();

    void handleNewClient();
    void handleClientReceive(int fd);
    void handleClientDisconnect(ClientMap::iterator &iter);
    void handleClientCommand(ClientMap::iterator &iter,
                             const char *buffer, size_t len);

    void handleClientCommand_help(ClientMap::iterator &iter,
                                  const std::string &argument);
    void handleClientCommand_who(ClientMap::iterator &iter,
                                 const std::string &argument);
    void handleClientCommand_setnick(ClientMap::iterator &iter,
                                     const std::string &argument);
    void handleClientCommand_unknown(ClientMap::iterator &iter,
                                     const std::string &argument);

    void handleClientMessage(ClientMap::iterator &iter,
                             const char *buffer, size_t len);

    void sendToClient(ClientMap::iterator &iter,
                      const void *buffer, size_t len);
    void sendToAllClients(const void *buffer, size_t len);

    void removeClient(ClientMap::iterator iter);

    static sig_atomic_t m_shutdown;
    Poller m_poller;
    TCPListener m_server;
    ClientMap m_clients;
};

sig_atomic_t Server::m_shutdown = 0;


//----------------------------------------------------------------------------
Server::Server() : m_poller(this)
{
    m_server.setReuseAddress();
    m_server.bind(InetAddress("0.0.0.0", 10000));
    m_server.listen();
}

//----------------------------------------------------------------------------
Server::~Server()
{
}

//----------------------------------------------------------------------------
void Server::onTerm()
{
    std::cout << "onTerm() called" << std::endl;
    m_shutdown = 1;
}

//----------------------------------------------------------------------------
int Server::run()
{
    m_poller.add(m_server.getFD(), POLLIN);

    InetAddress addr;
    std::cout << "Server listening on "
              << m_server.getLocalAddress(addr) << std::endl;

    while (!m_shutdown)
    {
        // sleep(5);
        // D'oh! A signal catched here will be ignored
        // until poll() returns again.

        std::cout << "Calling poll()..." << std::endl;
        if (m_poller.poll() == -1)
        {
            std::cout << "poll() returned -1, errno = "
                      << strerror(errno) << std::endl;
        }
    }

    shutdown();

    return 0;
}


//----------------------------------------------------------------------------
void Server::shutdown()
{
    m_poller.remove(m_server.getFD());

    char message[] = "# Server is going down...\n";
    sendToAllClients(message, sizeof(message));

    for (ClientMap::iterator iter = m_clients.begin();
         iter != m_clients.end(); ++iter)
    {
        std::cout << "Closing connection to "
                  << iter->second.nick << std::endl;

        m_poller.remove(iter->second.connection->getFD());
        delete iter->second.connection;
    }

    m_clients.clear();
}

//----------------------------------------------------------------------------
void Server::onPoll(int fd, pollevent_t events)
{
    if (fd == m_server.getFD())
    {
        handleNewClient();
    }
    else
    {
        handleClientReceive(fd);
    }
}


//----------------------------------------------------------------------------
void Server::handleNewClient()
{
    try
    {
        ClientEntry e;

        InetAddress addr;
        e.connection = m_server.accept(&addr);
        e.nick = addr.toString();

        std::cout << "New connection from " << e.nick << std::endl;

        m_poller.add(e.connection->getFD(), POLLIN);
        m_clients.insert(std::make_pair(e.connection->getFD(), e));

        std::string message;
        message.append("# New connection from ").append(e.nick).append("\n");
        sendToAllClients(message.c_str(), message.length());
    }
    catch (SocketException &e)
    {
        std::cerr << "Server error: " << e << std::endl;
        exit(-1);
    }
}

//----------------------------------------------------------------------------
void Server::handleClientReceive(int fd)
{
    ClientMap::iterator iter = m_clients.find(fd);
    if (iter != m_clients.end())
    {
        ClientEntry &entry = iter->second;

        try
        {
            char buffer[1024];
            int rc = entry.connection->recv(buffer, sizeof(buffer));

            if (rc == -1 && errno == EINTR)  return;

            if (rc == 0)
            {
                handleClientDisconnect(iter);
            }
            else
            {
                // Strip trailing newlines -- "netcat compatibility" ;-)
                --rc;
                while (rc-- >= 0 && (buffer[rc]=='\r' || buffer[rc]=='\n'));
                ++rc;

                if (buffer[0] == '/')
                {
                    handleClientCommand(iter, buffer, rc);
                }
                else
                {
                    handleClientMessage(iter, buffer, rc);
                }
            }
        }
        catch (SocketException &e)
        {
            std::cerr << "Client error: " << e << std::endl;
            removeClient(iter);

            std::string message;
            message.append("# Connection to ").append(entry.nick)
                .append(" broken.\n");
            sendToAllClients(message.c_str(), message.length());
        }
    }
}

//----------------------------------------------------------------------------
void Server::removeClient(ClientMap::iterator iter)
{
    m_poller.remove(iter->first);
    delete iter->second.connection;
    m_clients.erase(iter);
}

//----------------------------------------------------------------------------
void Server::handleClientDisconnect(ClientMap::iterator &iter)
{
    ClientEntry &entry = iter->second;
    std::cout << "Disconnect from " << entry.nick << std::endl;

    std::string message;
    message.append("# ").append(entry.nick).append(" disconnected.\n");

    removeClient(iter);

    sendToAllClients(message.c_str(), message.length());
}

//----------------------------------------------------------------------------
void Server::handleClientCommand(ClientMap::iterator &iter,
                                 const char *buffer, size_t len)
{
    std::string message(buffer, len);
    std::string command;
    std::string argument;

    size_t space = message.find(' ');
    if (space == std::string::npos)
    {
        command = message;
    }
    else
    {
        command = message.substr(0, space);
        argument = message.substr(space+1);
    }

    if (command == "/help")
    {
        handleClientCommand_help(iter, argument);
    }
    else if (command == "/who")
    {
        handleClientCommand_who(iter, argument);
    }
    else if (command == "/setnick")
    {
        handleClientCommand_setnick(iter, argument);
    }
    else
    {
        handleClientCommand_unknown(iter, argument);
    }
}

//----------------------------------------------------------------------------
void Server::handleClientCommand_help(ClientMap::iterator &iter,
                                      const std::string &argument)
{
    std::string message("# Available commands:\n"
                        "# /help ... This help text.\n"
                        "# /who ... List all logged in nicks.\n"
                        "# /setnick <nick> ... Set a new nick.\n");
    sendToClient(iter, message.c_str(), message.length());
}

//----------------------------------------------------------------------------
void Server::handleClientCommand_who(ClientMap::iterator &iter,
                                     const std::string &argument)
{
    std::string message("# Nicks logged in into this chat: ");

    bool first = true;
    for (ClientMap::const_iterator iter2 = m_clients.begin();
         iter2 != m_clients.end(); ++iter2)
    {
        if (!first)
        {
            message.append(", ");
        }

        message.append(iter2->second.nick);
        first = false;
    }

    message.append("\n");

    sendToClient(iter, message.c_str(), message.length());
}

//----------------------------------------------------------------------------
void Server::handleClientCommand_setnick(ClientMap::iterator &iter,
                                         const std::string &argument)
{
    std::string message;
    message.append("# ").append(iter->second.nick)
        .append(" changed nickname to ").append(argument).append("\n");

    iter->second.nick = argument;

    sendToAllClients(message.c_str(), message.length());
}

//----------------------------------------------------------------------------
void Server::handleClientCommand_unknown(ClientMap::iterator &iter,
                                         const std::string &argument)
{
    std::string message("# Unknown command. Use /help for a list of valid commands.\n");
    sendToClient(iter, message.c_str(), message.length());
}


//----------------------------------------------------------------------------
void Server::handleClientMessage(ClientMap::iterator &iter,
                                 const char *buffer, size_t len)
{
    ClientEntry &entry = iter->second;
    std::cout << "Received message from " << entry.nick << std::endl;

    std::string message;
    message.append(entry.nick).append("> ").append(buffer, len).append("\n");
    sendToAllClients(message.c_str(), message.length());
}

//----------------------------------------------------------------------------
void Server::sendToClient(ClientMap::iterator &iter,
                          const void *buffer, size_t len)
{
    if (len == 0)  return;

    try
    {
        iter->second.connection->send(buffer, len);
    }
    catch (SocketException &e)
    {
        std::cerr << "Client error on send: " << e << std::endl;
        removeClient(iter);
    }
}

//----------------------------------------------------------------------------
void Server::sendToAllClients(const void *buffer, size_t len)
{
    if (len == 0)  return;

    ClientMap::iterator iter = m_clients.begin();
    while (iter != m_clients.end())
    {
        try
        {
            iter->second.connection->send(buffer, len);
            ++iter;
        }
        catch (SocketException &e)
        {
            std::cerr << "Client error on send: " << e << std::endl;
            removeClient(iter++);
        }
    }
}



//----------------------------------------------------------------------------
int main()
{
    try
    {
        Server s;
        return s.run();
    }
    catch (Exception &e)
    {
        std::cerr << "Server error: " << e << std::endl;
    }
}
