#include <unistd.h>

#include "Exception.h"
#include "SystemCalls.h"
#include "Socket.h"

//----------------------------------------------------------------------------
SocketImpl::SocketImpl(int fd)
{
    m_fd = fd;
}

//----------------------------------------------------------------------------
SocketImpl::SocketImpl(int domain, int type)
{
    m_fd = SystemCall::socket(domain, type, 0);
}

//----------------------------------------------------------------------------
SocketImpl::~SocketImpl()
{
    if (m_fd != -1)
    {
        SystemCall::close(m_fd);
    }
}

//----------------------------------------------------------------------------
void SocketImpl::bind(const Address &addr)
{
    SystemCall::bind(getFD(), addr);
}

//----------------------------------------------------------------------------
int SocketImpl::connect(const Address &addr)
{
    return SystemCall::connect(getFD(), addr);
}

//----------------------------------------------------------------------------
void SocketImpl::listen(int backlog)
{
    SystemCall::listen(getFD(), backlog);
}

//----------------------------------------------------------------------------
int SocketImpl::accept(Address *addr) const
{
    return SystemCall::accept(getFD(), addr);
}

//----------------------------------------------------------------------------
int SocketImpl::recv(void *buf, size_t len)
{
    return SystemCall::recv(getFD(), buf, len, 0);
}

//----------------------------------------------------------------------------
int SocketImpl::send(const void *buf, size_t len)
{
    return SystemCall::send(getFD(), buf, len, 0);
}

//----------------------------------------------------------------------------
int SocketImpl::recvfrom(void *buf, size_t len, Address *addr)
{
    return SystemCall::recvfrom(getFD(), buf, len, 0, addr);
}

//----------------------------------------------------------------------------
int SocketImpl::sendto(const void *buf, size_t len, const Address &addr)
{
    return SystemCall::sendto(getFD(), buf, len, 0, addr);
}

//----------------------------------------------------------------------------
void SocketImpl::getsockopt(int level, int optname,
                            void *optval, socklen_t *optlen) const
{
    SystemCall::getsockopt(getFD(), level, optname, optval, optlen);
}

//----------------------------------------------------------------------------
void SocketImpl::setsockopt(int level, int optname,
                            const void *optval, socklen_t optlen)
{
    SystemCall::setsockopt(getFD(), level, optname, optval, optlen);
}

//----------------------------------------------------------------------------
int SocketImpl::getintsockopt(int level, int optname) const
{
    int optval;
    socklen_t optlen = sizeof(optval);
    getsockopt(level, optname, &optval, &optlen);
    return optval;
}

//----------------------------------------------------------------------------
void SocketImpl::setintsockopt(int level, int optname, int optval)
{
    setsockopt(level, optname, &optval, sizeof(optval));
}

//----------------------------------------------------------------------------
void SocketImpl::setNonBlockingOperation(bool enable)
{
    int flags = SystemCall::fcntl(getFD(), F_GETFL);

    if (enable)
    {
        flags |= O_NONBLOCK;
    }
    else
    {
        flags &= ~O_NONBLOCK;
    }

    SystemCall::fcntl(getFD(), F_SETFL, flags);
}

//----------------------------------------------------------------------------
void SocketImpl::setReuseAddress()
{
    setintsockopt(SOL_SOCKET, SO_REUSEADDR, 1);
}

//----------------------------------------------------------------------------
Address &SocketImpl::getLocalAddress(Address &addr) const
{
    SystemCall::getsockname(getFD(), addr);
    return addr;
}

//----------------------------------------------------------------------------
Address &SocketImpl::getPeerAddress(Address &addr) const
{
    SystemCall::getpeername(getFD(), addr);
    return addr;
}



//----------------------------------------------------------------------------
SocketBase::SocketBase(int fd) : m_socket(fd)
{
}

//----------------------------------------------------------------------------
SocketBase::SocketBase(int domain, int type) : m_socket(domain, type)
{
}

//----------------------------------------------------------------------------
SocketBase::~SocketBase()
{
}

//----------------------------------------------------------------------------
int SocketBase::getFD() const
{
    return m_socket.getFD();
}

//----------------------------------------------------------------------------
void SocketBase::setNonBlockingOperation(bool enable)
{
    m_socket.setNonBlockingOperation(enable);
}



//----------------------------------------------------------------------------
SocketPair::SocketPair()
{
    SystemCall::socketpair(PF_UNIX, SOCK_STREAM, 0, m_fds);
}

//----------------------------------------------------------------------------
SocketPair::~SocketPair()
{
}


//----------------------------------------------------------------------------
int SocketPair::getFD1() const
{
    return m_fds[0];
}

//----------------------------------------------------------------------------
int SocketPair::getFD2() const
{
    return m_fds[1];
}


//----------------------------------------------------------------------------
int SocketPair::recv1(void *buf, size_t len) const
{
    return SystemCall::recv(getFD1(), buf, len, 0);
}

//----------------------------------------------------------------------------
int SocketPair::recv2(void *buf, size_t len) const
{
    return SystemCall::recv(getFD2(), buf, len, 0);
}

//----------------------------------------------------------------------------
int SocketPair::send1(const void *buf, size_t len) const
{
    return SystemCall::send(getFD1(), buf, len, 0);
}

//----------------------------------------------------------------------------
int SocketPair::send2(const void *buf, size_t len) const
{
    return SystemCall::send(getFD2(), buf, len, 0);
}
