#ifndef SOCKET_H
#define SOCKET_H

#include "Address.h"

//----------------------------------------------------------------------------
class SocketImpl
{
    //------------------------------------------------------------------------
    SocketImpl(const SocketImpl &other);
    SocketImpl &operator=(const SocketImpl &other);

  public:
    //------------------------------------------------------------------------
    SocketImpl(int fd);
    SocketImpl(int domain, int type);
    ~SocketImpl();

    //------------------------------------------------------------------------
    inline int getFD() const { return m_fd; }

    //------------------------------------------------------------------------
    void bind(const Address &addr);
    int connect(const Address &addr);
    void listen(int backlog);
    int accept(Address *addr) const;

    //------------------------------------------------------------------------
    int recv(void *buf, size_t len);
    int send(const void *buf, size_t len);
    int recvfrom(void *buf, size_t len, Address *addr);
    int sendto(const void *buf, size_t len, const Address &addr);

    //------------------------------------------------------------------------
    void getsockopt(int level, int optname,
                    void *optval, socklen_t *optlen) const;
    void setsockopt(int level, int optname,
                    const void *optval, socklen_t optlen);
    int getintsockopt(int level, int optname) const;
    void setintsockopt(int level, int optname, int optval);

    //------------------------------------------------------------------------
    void setNonBlockingOperation(bool enable);
    void setReuseAddress();

    //------------------------------------------------------------------------
    Address &getLocalAddress(Address &addr) const;
    Address &getPeerAddress(Address &addr) const;

private:
    //------------------------------------------------------------------------
    int m_fd;
};


//----------------------------------------------------------------------------
class SocketBase
{
    //------------------------------------------------------------------------
    SocketBase(const SocketBase &other);
    SocketBase &operator=(const SocketBase &other);

  public:
    //------------------------------------------------------------------------
    SocketBase(int fd);
    SocketBase(int domain, int type);
    virtual ~SocketBase();

    //------------------------------------------------------------------------
    int getFD() const;

    //------------------------------------------------------------------------
    void setNonBlockingOperation(bool enable);

  protected:
    //------------------------------------------------------------------------
    SocketImpl m_socket;
};


//----------------------------------------------------------------------------
template <int domain, int type, class Address>
class ConnectionT : public SocketBase
{
    //------------------------------------------------------------------------
    ConnectionT(const ConnectionT &other);
    ConnectionT &operator=(const ConnectionT &other);

  public:
    //------------------------------------------------------------------------
    ConnectionT() : SocketBase(domain, type) {}
    ConnectionT(int fd) : SocketBase(fd) {}
    ~ConnectionT() {}

    //------------------------------------------------------------------------
    inline void bind(const Address &addr)
    {
        m_socket.bind(addr);
    }

    inline int connect(const Address &addr)
    {
        return m_socket.connect(addr);
    }

    inline int recv(void *buf, size_t len)
    {
        return m_socket.recv(buf, len);
    }

    inline int send(const void *buf, size_t len)
    {
        return m_socket.send(buf, len);
    }

    //------------------------------------------------------------------------
    inline Address &getLocalAddress(Address &addr) const 
    {
        (void)m_socket.getLocalAddress(addr);
        return addr;
    }

    inline Address &getPeerAddress(Address &addr) const
    {
        (void)m_socket.getPeerAddress(addr);
        return addr;
    }
};

//----------------------------------------------------------------------------
typedef ConnectionT<PF_INET, SOCK_STREAM, InetAddress> TCPConnection;
typedef ConnectionT<PF_INET, SOCK_DGRAM, InetAddress> UDPConnection;

//----------------------------------------------------------------------------
typedef ConnectionT<PF_UNIX, SOCK_STREAM, UnixAddress> TCPUnixConnection;
typedef ConnectionT<PF_UNIX, SOCK_DGRAM, UnixAddress> UDPUnixConnection;


//----------------------------------------------------------------------------
template <int domain, class Address>
class TCPListenerT : public SocketBase
{
    //------------------------------------------------------------------------
    TCPListenerT(const TCPListenerT &other);
    TCPListenerT &operator=(const TCPListenerT &other);

    //------------------------------------------------------------------------
    typedef ConnectionT<domain, SOCK_STREAM, Address> Connection;

  public:

    //------------------------------------------------------------------------
    TCPListenerT() : SocketBase(domain, SOCK_STREAM) {}

    TCPListenerT(const Address &addr) : SocketBase(domain, SOCK_STREAM)
    {
        bind(addr);
        listen();
    }

    ~TCPListenerT() {}

    //------------------------------------------------------------------------
    inline void bind(const Address &addr)
    {
        m_socket.bind(addr);
    }

    inline void listen(int backlog = 16)
    {
        m_socket.listen(backlog);
    }

    Connection *accept(Address *addr = NULL) const
    {
        return new Connection(m_socket.accept(addr));
    }

    inline void setReuseAddress()
    {
        m_socket.setReuseAddress();
    }

    //------------------------------------------------------------------------
    Address &getLocalAddress(Address &addr) const
    {
        (void)m_socket.getLocalAddress(addr);
        return addr;
    }
};

//----------------------------------------------------------------------------
typedef TCPListenerT<PF_INET, InetAddress> TCPListener;
typedef TCPListenerT<PF_UNIX, UnixAddress> TCPUnixListener;


//----------------------------------------------------------------------------
template <int domain, class Address>
class UDPListenerT : public SocketBase
{
    //------------------------------------------------------------------------
    UDPListenerT(const UDPListenerT &other);
    UDPListenerT &operator=(const UDPListenerT &other);

  public:
    //------------------------------------------------------------------------
    UDPListenerT() : SocketBase(domain, SOCK_DGRAM) {}
    ~UDPListenerT() {}

    //------------------------------------------------------------------------
    inline void bind(const Address &addr)
    {
        m_socket.bind(addr);
    }

    inline int recvfrom(void* buf, size_t len, Address *addr = NULL)
    {
        return m_socket.recvfrom(buf, len, addr);
    }

    inline int sendto(const void *buf, size_t len, const Address &addr)
    {
        return m_socket.sendto(buf, len, addr);
    }

    //------------------------------------------------------------------------
    inline Address &getLocalAddress(Address &addr) const
    {
        (void)m_socket.getLocalAddress(addr);
        return addr;
    }
};

//----------------------------------------------------------------------------
typedef UDPListenerT<PF_INET, InetAddress> UDPListener;
typedef UDPListenerT<PF_UNIX, UnixAddress> UDPUnixListener;



//----------------------------------------------------------------------------
class SocketPair
{
    //------------------------------------------------------------------------
    SocketPair(const SocketPair &other);
    SocketPair &operator=(const SocketPair &other);

  public:
    //------------------------------------------------------------------------
    SocketPair();
    ~SocketPair();

    //------------------------------------------------------------------------
    int getFD1() const;
    int getFD2() const;

    //------------------------------------------------------------------------
    int recv1(void *buf, size_t len) const;
    int recv2(void *buf, size_t len) const;

    //------------------------------------------------------------------------
    int send1(const void *buf, size_t len) const;
    int send2(const void *buf, size_t len) const;

  private:
    //------------------------------------------------------------------------
    int m_fds[2];
};

#endif //SOCKET_H
