#include <unistd.h>
#include <fcntl.h>
#include <sys/wait.h>

#include "Exception.h"
#include "SystemCalls.h"

namespace SystemCall
{

//----------------------------------------------------------------------------
int accept(int s, Address *addr)
{
    socklen_t addrlen = addr ? addr->getSockLen() : 0;
    int rc = ::accept(s, addr ? addr->getSockAddr() : NULL,
                      addr ? &addrlen : NULL);
    if (SocketException::isError(rc))
    {
        throw SocketException("accept()");
    }

    return rc;
}

//----------------------------------------------------------------------------
void bind(int s, const Address &addr)
{
    int rc = ::bind(s, addr.getSockAddr(), addr.getSockLen());
    if (SocketException::isError(rc))
    {
        throw SocketException("bind()");
    }
}

//----------------------------------------------------------------------------
void close(int s)
{
    int rc = ::close(s);
    if (SocketException::isError(rc))
    {
        throw SocketException("close()");
    }
}

//----------------------------------------------------------------------------
int connect(int s, const Address &addr)
{
    int rc = ::connect(s, addr.getSockAddr(), addr.getSockLen());
    if (SocketException::isError(rc))
    {
        throw SocketException("connect()");
    }

    return rc;
}

//----------------------------------------------------------------------------
int fcntl(int fd, int cmd)
{
    int rc = ::fcntl(fd, cmd);
    if (SocketException::isError(rc))
    {
        throw SocketException("fcntl()");
    }
    return rc;
}

//----------------------------------------------------------------------------
int fcntl(int fd, int cmd, long arg)
{
    int rc = ::fcntl(fd, cmd, arg);
    if (SocketException::isError(rc))
    {
        throw SocketException("fcntl()");
    }
    return rc;
}

//----------------------------------------------------------------------------
pid_t fork()
{
    pid_t rc = ::fork();
    if (SystemCallException::isError(rc))
    {
        throw SystemCallException("fork()");
    }
    return rc;
}


//----------------------------------------------------------------------------
struct hostent *gethostbyname(const char *name)
{
    struct hostent *host = ::gethostbyname(name);
    if (host == NULL)
    {
        throw GetHostByNameException(name);
    }
    return host;
}

//----------------------------------------------------------------------------
void getpeername(int s, Address &addr)
{
    socklen_t addrlen = addr.getSockLen();
    int rc = ::getpeername(s, addr.getSockAddr(), &addrlen);
    if (SocketException::isError(rc))
    {
        throw SocketException("getpeername()");
    }
}

//----------------------------------------------------------------------------
void getsockname(int s, Address &addr)
{
    socklen_t addrlen = addr.getSockLen();
    int rc = ::getsockname(s, addr.getSockAddr(), &addrlen);
    if (SocketException::isError(rc))
    {
        throw SocketException("getsockname()");
    }
}

//----------------------------------------------------------------------------
void getsockopt(int s, int level, int optname, void *optval, socklen_t *optlen)
{
    int rc = ::getsockopt(s, level, optname, optval, optlen);
    if (SocketException::isError(rc))
    {
        throw SocketException("getsockopt()");
    }
}

//----------------------------------------------------------------------------
void listen(int s, int backlog)
{
    int rc = ::listen(s, backlog);
    if (SocketException::isError(rc))
    {
        throw SocketException("listen()");
    }
}

//----------------------------------------------------------------------------
int recv(int s, void *buf, size_t len, int flags)
{
    int rc = ::recv(s, buf, len, flags);
    if (SocketException::isError(rc))
    {
        throw SocketException("recv()");
    }

    return rc;
}

//----------------------------------------------------------------------------
int recvfrom(int s, void *buf, size_t len, int flags, Address *addr)
{
    socklen_t addrlen = addr ? addr->getSockLen() : 0;
    int rc = ::recvfrom(s, buf, len, flags,
                        addr ? addr->getSockAddr() : NULL, &addrlen);
    if (SocketException::isError(rc))
    {
        throw SocketException("recvfrom()");
    }

    return rc;
}

//----------------------------------------------------------------------------
int recvmsg(int s, struct msghdr *msg, int flags)
{
    int rc = ::recvmsg(s, msg, flags);
    if (SocketException::isError(rc))
    {
        throw SocketException("recvmsg()");
    }

    return rc;
}

//----------------------------------------------------------------------------
int send(int s, const void *buf, size_t len, int flags)
{
    int rc = ::send(s, buf, len, flags);
    if (SocketException::isError(rc))
    {
        throw SocketException("send()");
    }

    return rc;
}

//----------------------------------------------------------------------------
int sendto(int s, const void *buf, size_t len, int flags, const Address &addr)
{
    int rc = ::sendto(s, buf, len, flags,
                      addr.getSockAddr(), addr.getSockLen());
    if (SocketException::isError(rc))
    {
        throw SocketException("sendto()");
    }

    return rc;
}

//----------------------------------------------------------------------------
int sendmsg(int s, const struct msghdr *msg, int flags)
{
    int rc = ::sendmsg(s, msg, flags);
    if (SocketException::isError(rc))
    {
        throw SocketException("sendmsg()");
    }

    return rc;
}

//----------------------------------------------------------------------------
void setsockopt(int s, int level, int optname,
                const void *optval, socklen_t optlen)
{
    int rc = ::setsockopt(s, level, optname, optval, optlen);
    if (SocketException::isError(rc))
    {
        throw SocketException("setsockopt()");
    }
}

//----------------------------------------------------------------------------
int socket(int domain, int type, int protocol)
{
    int rc = ::socket(domain, type, protocol);
    if (SocketException::isError(rc))
    {
        throw SocketException("socket()");
    }

    return rc;
}

//----------------------------------------------------------------------------
void socketpair(int d, int type, int protocol, int *sv)
{
    int rc = ::socketpair(d, type, protocol, sv);
    if (SocketException::isError(rc))
    {
        throw SocketException("socketpair()");
    }
}

//----------------------------------------------------------------------------
pid_t wait(int *status)
{
    pid_t rc = ::wait(status);
    if (SystemCallException::isError(rc))
    {
        throw SystemCallException("wait()");
    }

    return rc;
}

} // namespace SystemCall
